﻿' Description:    A bound image listbox with various painting options
' Date Created:   November 2009
' Author:         Kevin Carbis
' Author Email:   kcarbis@comcast.net / kevin@staffinder.net
' Websites:       www.staffinder.net/carbisoft
' Credits:        I took the RoundRectangle function from www.menendezpoo.com.
'                 His was simpler than my version
' License:        If you use this control in a comertial application you need to link back to my site.
' Version:        1.0   Initial Release
'                 1.1   Added SelectedForeColor and SelectedFont properties.
'                       Fixed bug in Paint event that cause error when using the control in unbound mode.
'                       Added ability to support binary or image database fields when in bound mode.
'                       Control will now dither all items when the control is disabled
'                       Fixed the "Listbox flicker" problem when scrolling or resizing

Imports System.Drawing.Drawing2D
Imports System.IO
Imports System.ComponentModel
Imports System.Runtime.InteropServices


Public Class ImageListBox
   Inherits System.Windows.Forms.ListBox

   Private _Image As Bitmap
   Private _ImageMember As String = ""
   Private _ImageSize As Size = New Size(16, 16)
   Private _RowDividerVisible As Boolean = True
   Private _RowDividerWidth As Integer = 1
   Private _RowDividerColor As Color = Color.FromArgb(194, 217, 240)
   Private _RowDividerColorGradient As Color = Color.FromArgb(227, 235, 246)
   Private _SelectedBackColor As Color = Color.FromArgb(255, 242, 199)
   Private _SelectedBorderColor As Color = Color.Orange
   Private _SelectedCornerRadius As Integer = 5
   Private _SelectedForeColor As Color = Color.FromKnownColor(KnownColor.WindowText)
   Private _SelectedFont As Font = Me.Font
   Private _CursorPos As System.Drawing.Point
   Private _ScrollViewer

   Private Const WM_ERASEBKGND As Integer = &H14
   Private Const WM_MOUSEWHEEL As Integer = &H20A
   Private Const WM_VSCROLL = &H115
   Private Const SB_LINEUP = 0
   Private Const SB_LINEDOWN = 1
   Private Const WM_KEYDOWN As Integer = &H100
   Private Const WM_KEYUP As Integer = &H101
   Private Const LB_ADDSTRING As Integer = &H180
   Private Const LB_INSERTSTRING As Integer = &H181
   Private Const LB_DELETESTRING As Integer = &H182
   Private Const LB_RESETCONTENT As Integer = &H184
   <DllImport("user32.dll", CharSet:=CharSet.Auto, SetLastError:=False)> _
   Private Shared Function SendMessage(ByVal hWnd As IntPtr, ByVal Msg As Int32, ByVal wParam As IntPtr, ByVal lParam As IntPtr) As IntPtr
   End Function

   Public Sub New()
      SetStyle(ControlStyles.OptimizedDoubleBuffer, True)
      SetStyle(ControlStyles.AllPaintingInWmPaint, True)
      SetStyle(ControlStyles.ResizeRedraw, True)
      SetStyle(ControlStyles.UserPaint, True)
      DrawMode = Windows.Forms.DrawMode.OwnerDrawFixed
      ItemHeight = 22
   End Sub

   <EditorBrowsable(EditorBrowsableState.Always)> _
   <Category("Appearance"), Description("Gets or sets the visibility of the row divider.")> _
   <DefaultValue(True)> _
   Public Property RowDividerVisible() As Boolean
      Get
         Return _RowDividerVisible
      End Get
      Set(ByVal value As Boolean)
         _RowDividerVisible = value
         Invalidate()
      End Set
   End Property
   ''' <summary>Gets or sets the row divider width.</summary>
   <EditorBrowsable(EditorBrowsableState.Always)> _
   <Category("Appearance"), Description("Gets or sets the row divider width.")> _
   <DefaultValue(1)> _
   Public Property RowDividerWidth() As Integer
      Get
         Return _RowDividerWidth
      End Get
      Set(ByVal value As Integer)
         If value < 1 Or value > 10 Then
            Throw New ArgumentOutOfRangeException("Width", "Invalid Property Assignment")
         End If
         _RowDividerWidth = value
         Invalidate()
      End Set
   End Property
   ''' <summary>Gets or sets the row divider color.</summary>
   <EditorBrowsable(EditorBrowsableState.Always)> _
   <Category("Appearance"), Description("Gets or sets the row divider color.")> _
   Public Property RowDividerColor() As Color
      Get
         Return _RowDividerColor
      End Get
      Set(ByVal value As Color)
         _RowDividerColor = value
         Invalidate()
      End Set
   End Property
   ''' <summary>Gets or sets the row divider gradient.</summary>
   <Category("Appearance"), Description("Gets or sets the row divider gradient.")> _
   Public Property RowDividerColorGradient() As Color
      Get
         Return _RowDividerColorGradient
      End Get
      Set(ByVal value As Color)
         _RowDividerColorGradient = value
         Invalidate()
      End Set
   End Property
   ''' <summary>Gets or sets the back color of the selected row.</summary>
   <Category("Appearance"), Description("Gets or sets the back color of the selected row.")> _
   <DefaultValue(GetType(Color), "255, 242, 199")> _
   Public Property SelectedBackColor() As Color
      Get
         Return _SelectedBackColor
      End Get
      Set(ByVal value As Color)
         _SelectedBackColor = value
      End Set
   End Property
   ''' <summary>Gets or sets the color of the border.</summary>
   <Category("Appearance"), Description("Gets or sets the color of the border.")> _
   <DefaultValue(GetType(Color), "Orange")> _
   Public Property SelectedBorderColor() As Color
      Get
         Return _SelectedBorderColor
      End Get
      Set(ByVal value As Color)
         _SelectedBorderColor = value
      End Set
   End Property
   ''' <summary>Gets or sets the size of the corner radius.</summary>
   <Category("Appearance"), Description("Gets or sets the size of the corner radius.")> _
   <DefaultValue(5)> _
   Public Property SelectedCornerRadius() As Integer
      Get
         Return _SelectedCornerRadius
      End Get
      Set(ByVal value As Integer)
         If value < 0 Then
            _SelectedCornerRadius = 0
         ElseIf value > 10 Then
            _SelectedCornerRadius = 10
         Else
            _SelectedCornerRadius = value
         End If
      End Set
   End Property
   ''' <summary>Gets or sets the color of the selected rows text.</summary>
   <Category("Appearance"), Description("Gets or sets the color of the selected rows text.")> _
   <DefaultValue(GetType(Color), "Black")> _
   Public Property SelectedForeColor() As Color
      Get
         Return _SelectedForeColor
      End Get
      Set(ByVal value As Color)
         _SelectedForeColor = value
         Invalidate()
      End Set
   End Property
   ''' <summary>Gets or sets the font style of the selected row.</summary>
   <Category("Appearance"), Description("Gets or sets the font style of the selected row.")> _
   <DefaultValue(GetType(Font), "Microsoft Sans Serif, 8.25pt")> _
   Public Property SelectedFont() As Font
      Get
         Return _SelectedFont
      End Get
      Set(ByVal value As Font)
         _SelectedFont = value
         Invalidate()
      End Set
   End Property
   ''' <summary>Gets or sets the name of the database image field for data binding.</summary>
   <Category("Data"), Description("Gets or sets the name of the database image field for data binding.")> _
   Public Property ImageMember() As String
      Get
         Return _ImageMember
      End Get
      Set(ByVal value As String)
         _ImageMember = value
      End Set
   End Property
   ''' <summary>Gets or sets the size of the scaled image. If the image differs it will be scaled to this size.</summary>
   <Category("Appearance"), Description("Gets or sets the size of the scaled image. If the image differs it will be scaled to this size.")> _
   Public Property ImageSize() As Size
      Get
         Return _ImageSize
      End Get
      Set(ByVal value As Size)
         _ImageSize = value
      End Set
   End Property

   Protected Overrides Sub OnPaint(ByVal e As System.Windows.Forms.PaintEventArgs)
      'to stop the flickering when scrolling or resizing we needed to stop the paintbackground messages.
      'however, this caused the background to not paint in several situations when we actually need it to.
      'therefore, we are forced to do all painting on our own so double buffering is truely used.

      'fill the entire background of the control
      e.Graphics.FillRectangle(New SolidBrush(Me.BackColor), ClientRectangle)

      'now only paint the items that need to be.
      For i As Integer = 0 To Items.Count - 1 Step 1
         Dim rec As Rectangle = Me.GetItemRectangle(i)
         If e.ClipRectangle.IntersectsWith(rec) Then
            If SelectedIndices.Contains(i) Then
               OnDrawItem(New DrawItemEventArgs(e.Graphics, SelectedFont, rec, i, DrawItemState.Selected))
            Else
               OnDrawItem(New DrawItemEventArgs(e.Graphics, Font, rec, i, DrawItemState.Default))
            End If
         End If
      Next
   End Sub
   Private Sub ImageListBox_DrawItem(ByVal sender As Object, ByVal e As System.Windows.Forms.DrawItemEventArgs) Handles Me.DrawItem
      Try
         If e.Index > -1 Then
            'clear out the canvas
            e.Graphics.SmoothingMode = Drawing2D.SmoothingMode.AntiAlias
            e.Graphics.FillRectangle(New SolidBrush(Me.BackColor), e.Bounds)

            'draw the background depending on if its a selected row or not
            Dim imgWidth As Integer = 0
            Dim innerRec As New Rectangle(e.Bounds.X + 1, e.Bounds.Y + 1, e.Bounds.Width - 2, e.Bounds.Height - 2)
            If (e.State And DrawItemState.Selected) = DrawItemState.Selected And Enabled Then
               Using path As GraphicsPath = RoundRectangle(innerRec, SelectedCornerRadius, Corners.All)
                  e.Graphics.FillPath(New SolidBrush(SelectedBackColor), path)
                  Using p As Pen = New Pen(SelectedBorderColor, 1)
                     e.Graphics.DrawPath(p, path)
                  End Using
               End Using
            ElseIf (e.State And DrawItemState.Disabled) = DrawItemState.Disabled Then
            Else
               If _RowDividerVisible Then
                  Dim b As New Drawing2D.LinearGradientBrush(innerRec, _RowDividerColor, _RowDividerColorGradient, Drawing2D.LinearGradientMode.Horizontal)
                  e.Graphics.DrawLine(New Pen(b, _RowDividerWidth), innerRec.X, innerRec.Bottom - (_RowDividerWidth \ 2), innerRec.Right, innerRec.Bottom - (_RowDividerWidth \ 2))
               End If
            End If
            Dim LeftMargin As Integer = 5

            'draw the image if we have one
            If ImageMember <> "" Then
               'got an image to draw here
               If Not IsDBNull(CType(Me.Items(e.Index), DataRowView)(ImageMember)) Then
                  'determine column type
                  Dim itmImage As Bitmap
                  If CType(Me.Items(e.Index), DataRowView)(ImageMember).GetType Is GetType(Bitmap) Then
                     itmImage = CType(Me.Items(e.Index), DataRowView)(ImageMember)
                  Else
                     itmImage = BytesToImage(CType(Me.Items(e.Index), DataRowView)(ImageMember))
                  End If

                  Dim ScaleImage As Bitmap = ResizeImage(itmImage, ImageSize.Width, ImageSize.Height)
                  Dim intTop As Integer = (ItemHeight - ScaleImage.Height) \ 2
                  e.Graphics.DrawImage(ScaleImage, innerRec.X + LeftMargin, innerRec.Y + ((innerRec.Height - ScaleImage.Height) \ 2), ScaleImage.Width, ScaleImage.Height)
               End If
               LeftMargin += ImageSize.Width + 5
            End If

            'draw the text
            Dim strText As String = ""
            If Items.Count = 0 And Site IsNot Nothing AndAlso Site.DesignMode Then
               strText = Me.Name
            ElseIf Items(e.Index).GetType Is GetType(DataRowView) Then
               strText = CType(Me.Items(e.Index), DataRowView)(Me.DisplayMember)
            Else
               strText = Items(e.Index).ToString
            End If

            Dim textRec As RectangleF = innerRec
            textRec.X += LeftMargin
            textRec.Width -= LeftMargin

            Dim sf As New StringFormat
            sf.Alignment = StringAlignment.Near
            sf.LineAlignment = StringAlignment.Center
            sf.Trimming = StringTrimming.EllipsisWord
            sf.FormatFlags = StringFormatFlags.NoWrap
            If Enabled = False Then
               e.Graphics.DrawString(strText, Font, New Drawing.SolidBrush(Color.FromArgb(90, ForeColor)), textRec, sf)
            ElseIf (e.State And DrawItemState.Selected) = DrawItemState.Selected Then
               e.Graphics.DrawString(strText, SelectedFont, New Drawing.SolidBrush(SelectedForeColor), textRec, sf)
            Else
               e.Graphics.DrawString(strText, Font, New Drawing.SolidBrush(ForeColor), textRec, sf)
            End If
         End If
      Catch ex As Exception
         'Console.WriteLine(ex.Message)
      End Try
   End Sub

   Private Sub ImageListBox_MeasureItem(ByVal sender As Object, ByVal e As System.Windows.Forms.MeasureItemEventArgs) Handles Me.MeasureItem
      e.ItemHeight = ItemHeight
   End Sub

   Private Function RoundRectangle(ByVal r As Rectangle, ByVal radius As Integer, ByVal corners As Corners) As GraphicsPath
      Dim path As GraphicsPath = New GraphicsPath()
      If r.Width <= 0 Or r.Height <= 0 Then Return path

      Dim d As Integer = radius * 2

      Dim nw As Integer = IIf((corners And corners.NorthWest) = corners.NorthWest, d, 0)
      Dim ne As Integer = IIf((corners And corners.NorthEast) = corners.NorthEast, d, 0)
      Dim se As Integer = IIf((corners And corners.SouthEast) = corners.SouthEast, d, 0)
      Dim sw As Integer = IIf((corners And corners.SouthWest) = corners.SouthWest, d, 0)

      path.AddLine(r.Left + nw, r.Top, r.Right - ne, r.Top)

      If ne > 0 Then
         path.AddArc(Rectangle.FromLTRB(r.Right - ne, r.Top, r.Right, r.Top + ne), -90, 90)
      End If

      path.AddLine(r.Right, r.Top + ne, r.Right, r.Bottom - se)

      If se > 0 Then
         path.AddArc(Rectangle.FromLTRB(r.Right - se, r.Bottom - se, r.Right, r.Bottom), 0, 90)
      End If

      path.AddLine(r.Right - se, r.Bottom, r.Left + sw, r.Bottom)

      If sw > 0 Then
         path.AddArc(Rectangle.FromLTRB(r.Left, r.Bottom - sw, r.Left + sw, r.Bottom), 90, 90)
      End If

      path.AddLine(r.Left, r.Bottom - sw, r.Left, r.Top + nw)

      If nw > 0 Then
         path.AddArc(Rectangle.FromLTRB(r.Left, r.Top, r.Left + nw, r.Top + nw), 180, 90)
      End If

      path.CloseFigure()
      Return path
   End Function
   Private Enum Corners
      None = 0
      NorthWest = 2
      NorthEast = 4
      SouthEast = 8
      SouthWest = 16
      All = NorthWest Or NorthEast Or SouthEast Or SouthWest
      North = NorthWest Or NorthEast
      South = SouthEast Or SouthWest
      East = NorthEast Or SouthEast
      West = NorthWest Or SouthWest
   End Enum
   Protected Overloads Overrides Sub WndProc(ByRef m As Message)
      'this fixes the weird mousewheel problem when scrolling by converting the mousewheel message
      'to a standard vertical scroll message
      'I found it on the web at http://www.dnzone.com/go?1328 but modified it slightly to ignore the 
      'erase background message
      Select Case m.Msg
         Case CInt(WM_ERASEBKGND)
            'always ignore the background paint message
            Exit Sub
         Case CInt(WM_MOUSEWHEEL)
            Dim X As Integer
            Dim Y As Integer
            Dim byts As Byte() = System.BitConverter.GetBytes(CInt(m.WParam))
            X = BitConverter.ToInt16(byts, 0)
            Y = BitConverter.ToInt16(byts, 2)
            If Y > 0 Then
               If TopIndex = 0 Then Exit Sub
               SendMessage(Me.Handle, CInt(WM_VSCROLL), CType(SB_LINEUP, IntPtr), IntPtr.Zero)
            Else
               If TopIndex + Me.ClientSize.Height / Me.ItemHeight >= Items.Count Then Exit Sub
               SendMessage(Me.Handle, CInt(WM_VSCROLL), CType(SB_LINEDOWN, IntPtr), IntPtr.Zero)
               Exit Sub
            End If
         Case CInt(WM_KEYDOWN)
            Invalidate()
         Case CInt(WM_VSCROLL)
            Invalidate()
         Case LB_ADDSTRING, LB_INSERTSTRING, LB_DELETESTRING, LB_RESETCONTENT
            Invalidate()
      End Select
      MyBase.WndProc(m)
   End Sub

   Private Function BytesToImage(ByVal bData As Byte()) As Image
      Dim sImage As New MemoryStream(bData)
      Return Image.FromStream(sImage)
   End Function
   Private Function ResizeImage(ByVal Img As Image, ByVal Width As Integer, ByVal Height As Integer) As Bitmap
      Try
         Dim Result As New Bitmap(Width, Height)
         Using g As Graphics = Graphics.FromImage(Result)
            g.DrawImage(Img, 0, 0, Width, Height)
         End Using
         Return Result
      Catch ex As Exception
         Return Nothing
      End Try
   End Function
End Class
